#include "../global.h"

Spectrum::Spectrum(Epp *e, ScreenOutput *disp, AudioOutput *au, Keyboard *kb,
                    Joystick *js, const char *sna) : Machine(e, disp, au) {

    keyboard=kb;
    joystick=js;
    snapshot=sna;

    sndflag=false;
    sndpos=0;

    // Prevent a noise on startup

    for(int i=0; i<44100/50; i++) {

        sample[i]=0;
    }
}

void Spectrum::createComponents(void) {

    cpu = new Z80(this);
    vid = new SinclairVideo(this, display, NULL);
}

void Spectrum::destroyComponents(void) {

    delete cpu;
    delete vid;
}

void Spectrum::readROMs(void) {

    // Load the ROM - this is write protected in memory

    readROM("roms/spectrum.rom", memory, 16384);
}

void Spectrum::readMirage(const char *snapshot) {

    // Reads a Mirage (.sna) format snapshot file

    uint8 IFF=0;
    uint16 SP=0, PC=0;
    uint8 header[27];

    ifstream input(snapshot, ios::in|ios::bin);
    input.read(header, 27);
    input.read(memory+16384, 48*1024);

    if (input.gcount()!=48*1024) {

        Epp::log("Error: Bad snapshot file!");
        exit(1);
    }

    cpu->setRegister(Z80::I, header[0]);
    cpu->setRegister(Z80::L1, header[1]);
    cpu->setRegister(Z80::H1, header[2]);
    cpu->setRegister(Z80::E1, header[3]);
    cpu->setRegister(Z80::D1, header[4]);
    cpu->setRegister(Z80::C1, header[5]);
    cpu->setRegister(Z80::B1, header[6]);
    cpu->setRegister(Z80::F1, header[7]);
    cpu->setRegister(Z80::A1, header[8]);
    cpu->setRegister(Z80::L, header[9]);
    cpu->setRegister(Z80::H, header[10]);
    cpu->setRegister(Z80::E, header[11]);
    cpu->setRegister(Z80::D, header[12]);
    cpu->setRegister(Z80::C, header[13]);
    cpu->setRegister(Z80::B, header[14]);
    cpu->setRegister(Z80::IYl, header[15]);
    cpu->setRegister(Z80::IYh, header[16]);
    cpu->setRegister(Z80::IXl, header[17]);
    cpu->setRegister(Z80::IXh, header[18]);
    if (header[19] & 0x04) IFF |= 0x01;

    cpu->setRegister(Z80::F, header[21]);
    cpu->setRegister(Z80::A, header[22]);

    SP |= (header[23] << 0);
    SP |= (header[24] << 8);
    PC |= readMem(cpu, SP) << 0;
    SP++;
    PC |= readMem(cpu, SP) << 8;
    SP++;

    cpu->setRegister(Z80::SPh, (byte)(SP >> 8));
    cpu->setRegister(Z80::SPl, (byte)(SP & 0xFF));
    cpu->setRegister(Z80::PCh, (byte)(PC >> 8));
    cpu->setRegister(Z80::PCl, (byte)(PC &0xFF));

    if (header[25]==1)
        IFF |= 0x02;
    else if (header[25]==2)
        IFF |= 0x04;

    cpu->setRegister(Z80::IFF, IFF);
    vid->setBorder(header[26]);
}

void Spectrum::reset() {

    cpu->reset();
}

void Spectrum::emulate() {


    readROMs();
    createComponents();

    // Load the snapshot is one is given

    if (snapshot) {

        readMirage(snapshot);
    }

    vid->setupDisplay();

    // One frame's worth of sample

    void *synth=audio->createSynth(sample, 44100/50, 44100);
    void *chan=audio->openChannel(255, 127);

    // Set CPU defaults

    cpu->setRunTime(69888);

    // Main loop

    for(;;) {

        for(int fl=0; fl<12; fl++) {

            joystick->update();
            readInput();

            // Run the CPU

            cpu->execute();

            // If the audio is high, then fill to the end of the sample
            // To avoid "clicking".

            if(sndflag) {

                for(int i=sndpos; i<=44100/50; i++) {

                    sample[i]=127;
                }
            }

            // We now have a wave to play, so load it into sound
            // card, and play it.

            audio->modifySynth(synth, sample);
            audio->playWaveform(synth, chan);

            // Now empty the waveform array

            for(int i=0; i<44100/50; i++) {

                sample[i]=0;
            }
            sndpos=0;

            // What int vector does the Speccy generate?

            cpu->interrupt(0xffff);

            // Update the display

            vid->update();
            audio->update();

            if(epp->checkInterface()) {

                destroyComponents();

                audio->stopChannel(chan);
                audio->closeChannel(chan);

                audio->freeSynth(synth);
                return;
            }
        }

        // Invert hardware flash every few frames

        vid->invert();
    }
}

void Spectrum::readInput(void) {

    keyvalue[0]=0xFF;

    if(keyboard->keyPressed(Keyboard::k_lshift)) keyvalue[0] &= 0xFE;
    if(keyboard->keyPressed(Keyboard::k_z)) keyvalue[0] &= 0xFD;
    if(keyboard->keyPressed(Keyboard::k_x)) keyvalue[0] &= 0xFB;
    if(keyboard->keyPressed(Keyboard::k_c)) keyvalue[0] &= 0xF7;
    if(keyboard->keyPressed(Keyboard::k_v)) keyvalue[0] &= 0xEF;

    keyvalue[1]=0xFF;

    if(keyboard->keyPressed(Keyboard::k_a)) keyvalue[1] &= 0xFE;
    if(keyboard->keyPressed(Keyboard::k_s)) keyvalue[1] &= 0xFD;
    if(keyboard->keyPressed(Keyboard::k_d)) keyvalue[1] &= 0xFB;
    if(keyboard->keyPressed(Keyboard::k_f)) keyvalue[1] &= 0xF7;
    if(keyboard->keyPressed(Keyboard::k_g)) keyvalue[1] &= 0xEF;

    keyvalue[2]=0xFF;

    if(keyboard->keyPressed(Keyboard::k_q)) keyvalue[2] &= 0xFE;
    if(keyboard->keyPressed(Keyboard::k_w)) keyvalue[2] &= 0xFD;
    if(keyboard->keyPressed(Keyboard::k_e)) keyvalue[2] &= 0xFB;
    if(keyboard->keyPressed(Keyboard::k_r)) keyvalue[2] &= 0xF7;
    if(keyboard->keyPressed(Keyboard::k_t)) keyvalue[2] &= 0xEF;

    keyvalue[3]=0xFF;

    if(keyboard->keyPressed(Keyboard::k_1)) keyvalue[3] &= 0xFE;
    if(keyboard->keyPressed(Keyboard::k_2)) keyvalue[3] &= 0xFD;
    if(keyboard->keyPressed(Keyboard::k_3)) keyvalue[3] &= 0xFB;
    if(keyboard->keyPressed(Keyboard::k_4)) keyvalue[3] &= 0xF7;
    if(keyboard->keyPressed(Keyboard::k_5)) keyvalue[3] &= 0xEF;

    keyvalue[4]=0xFF;

    if(keyboard->keyPressed(Keyboard::k_0)) keyvalue[4] &= 0xFE;
    if(keyboard->keyPressed(Keyboard::k_9)) keyvalue[4] &= 0xFD;
    if(keyboard->keyPressed(Keyboard::k_8)) keyvalue[4] &= 0xFB;
    if(keyboard->keyPressed(Keyboard::k_7)) keyvalue[4] &= 0xF7;
    if(keyboard->keyPressed(Keyboard::k_6)) keyvalue[4] &= 0xEF;

    keyvalue[5]=0xFF;

    if(keyboard->keyPressed(Keyboard::k_p)) keyvalue[5] &= 0xFE;
    if(keyboard->keyPressed(Keyboard::k_o)) keyvalue[5] &= 0xFD;
    if(keyboard->keyPressed(Keyboard::k_i)) keyvalue[5] &= 0xFB;
    if(keyboard->keyPressed(Keyboard::k_u)) keyvalue[5] &= 0xF7;
    if(keyboard->keyPressed(Keyboard::k_y)) keyvalue[5] &= 0xEF;

    keyvalue[6]=0xFF;

    if(keyboard->keyPressed(Keyboard::k_enter)) keyvalue[6] &= 0xFE;
    if(keyboard->keyPressed(Keyboard::k_l)) keyvalue[6] &= 0xFD;
    if(keyboard->keyPressed(Keyboard::k_k)) keyvalue[6] &= 0xFB;
    if(keyboard->keyPressed(Keyboard::k_j)) keyvalue[6] &= 0xF7;
    if(keyboard->keyPressed(Keyboard::k_h)) keyvalue[6] &= 0xEF;

    keyvalue[7]=0xFF;

    if(keyboard->keyPressed(Keyboard::k_spc)) keyvalue[7] &= 0xFE;
    if(keyboard->keyPressed(Keyboard::k_rshift)) keyvalue[7] &= 0xFD;
    if(keyboard->keyPressed(Keyboard::k_m)) keyvalue[7] &= 0xFB;
    if(keyboard->keyPressed(Keyboard::k_n)) keyvalue[7] &= 0xF7;
    if(keyboard->keyPressed(Keyboard::k_b)) keyvalue[7] &= 0xEF;

    stickvalue=0;

    if(joystick->jright()) stickvalue |= 0x01;
    if(joystick->jleft()) stickvalue |= 0x02;
    if(joystick->jdown()) stickvalue |= 0x04;
    if(joystick->jup()) stickvalue |= 0x08;
    if(joystick->jfire1()) stickvalue |= 0x10;

}

uint8 Spectrum::readMem(const Component *comp, const uint16 addr) {

    return memory[addr];
}

void Spectrum::writeMem(const Component *comp, const uint16 addr, const uint8 data) {

    // ROM is write protected
    
    if (addr>=23296) {

        memory[addr]=data;
    }
    
    // Video RAM
    else if(addr>=16384) {

        if (memory[addr]!=data) {

            memory[addr]=data;
            vid->writeVideoRAM(addr-16384, data);
        }
    }
}

uint8 Spectrum::in(const Component *comp, const uint16 port) {

    // Joystick
    if ((port & 0xFF) == 31) return stickvalue;

    // Any other odd address, then exit.
    if (port & 0x01) return 0;

    uint8 data=0xFF;
    
    uint8 top = ~((port & 0xFF00) >> 8);
    
    if(top & 0x01) data &= keyvalue[0];
    if(top & 0x02) data &= keyvalue[1];
    if(top & 0x04) data &= keyvalue[2];
    if(top & 0x08) data &= keyvalue[3];
    if(top & 0x10) data &= keyvalue[4];
    if(top & 0x20) data &= keyvalue[5];
    if(top & 0x40) data &= keyvalue[6];
    if(top & 0x80) data &= keyvalue[7];

    return data;
}

void Spectrum::out(const Component *comp, const uint16 port, const uint8 data) {

    // I only know about even ports (ULA)
    if(port & 0x01) return;

    // Check for audio out
    
    if((!sndflag) && (data & 0x10)) {

        // We divide down the CPU time to get the position within the sample
        
        sndpos=cpu->getTime()/80;
        sndflag=true;
    }
    else if((sndflag) && !(data & 0x10)) {

        // If the audio goes low, fill the sample to this point with
        // high data.
        for(uint16 i=sndpos; i<=cpu->getTime()/80; i++) {

            sample[i]=127;
        }
        sndflag=false;
    }

    // Set the border (setBorder() will strip the ms bits)
    vid->setBorder(data);
}

